# mime.tcl
#
# MIME message display.
#
# Copyright (c) 1993 Xerox Corporation.
# Use and copying of this software and preparation of derivative works based
# upon this software are permitted. Any distribution of this software or
# derivative works must comply with all applicable United States export
# control laws. This software is made available AS IS, and Xerox Corporation
# makes no warranty about the software, its performance or its conformity to
# any specification.

proc Mime_Init {} {
    global mime env
    if [info exists mime(init)] {
	return
    }
    # Make sure Metamail is on the path
    set hit 0
    foreach dir [split $env(PATH) :] {
	if {[string compare $dir $mime(dir)] == 0} {
	    set hit 1
	    break
	}
    }
    if {! $hit} {
	set env(PATH) $mime(dir):$env(PATH)
    }

    set mime(init) 1
    set mime(seed) 1
    set mime(junkfiles) {}
    set mime(showproc,text/plain) 		Mime_ShowText
    set mime(showproc,text/enriched)		Mime_ShowText
    set mime(showproc,text/richtext)		Mime_ShowText
    set mime(showproc,multipart) 		Mime_ShowMultipart
    set mime(showproc,application/octet-stream) Mime_ShowApplicationOctet
    set mime(showproc,application/postscript)   Mime_ShowDirect
    set mime(showproc,message/external-body)	Mime_ShowMessageExternal
    set mime(showproc,image) 			Mime_ShowDirect
    set mime(showproc,audio) 			Mime_ShowAudio

    set mime(showproc,default) 			Mime_ShowDefault
}
proc Mime_Cleanup {} {
    global mime
    set cmd rm
    foreach f $mime(junkfiles) {
	if [file exists $f] {
	    lappend cmd $f
	}
    }
    if {$cmd != "rm"} {
	eval exec $cmd
    }
    set mime(junkfiles) {}
}

proc Mime_ShowBody { tkw fileName contentType encoding} {
    global mime mimeHdr
    catch {unset mimeHdr}
    Exmh_Status "Formatting MIME message ..." red
    set mimeHdr(0,0,file) $fileName
    MimeHeader 0 0 $contentType $encoding
    MimeShow $tkw 0 0
}
proc MimeShow {tkw depth part} {
    global mime mimeHdr
    set fileName $mimeHdr($depth,$part,file)
    set type $mimeHdr($depth,$part,type)
    foreach t [list $type [file dirname $type] default] {
	if [info exists mime(showproc,$t)] {
	    if [$mime(showproc,$t) $tkw $depth $part] {
		exec rm $fileName
	    }
	    return
	}
    }
}
proc MimeHeader { depth part contentType encoding } {
    global mimeHdr
    # mimeHdr contains state about nested body parts:
    # mimeHdr($depth,$part,type)	Content-Type
    # mimeHdr($depth,$part,encoding)	Content-Transfer-Encoding
    # mimeHdr($depth,$part,file)	Tmp file containing body
    # mimeHdr($depth,$part,params)	Parameter names (e.g., boundary)
    # mimeHdr($depth,$part,param,$key)	Parameter value
    # mimeHdr($depth,$part,hdrs)	List of subpart mail headers
    # mimeHdr($depth,$part,hdr,$key)	Subpart mail header value

    set params [split $contentType \;]
    set type [string tolower [string trim [lindex $params 0]]]
    if {$type == "text"} {
	set type text/plain
    }
    set mimeHdr($depth,$part,hdr,content-type) $contentType
    set mimeHdr($depth,$part,type) $type
    set mimeHdr($depth,$part,encoding) $encoding
    set mimeHdr($depth,$part,params) {}
    foreach sub [lrange $params 1 end] {
	set sub [string trim $sub]
	if [regexp {([^=]+)=(.+)} $sub match key val] {
	    set key [string tolower $key]
	    set val [string trim $val \"]
	    lappend mimeHdr($depth,$part,params) $key
	    set mimeHdr($depth,$part,param,$key) $val
	}
    }
    return $type
}

proc Mime_ShowDefault { tkw depth part args} {
    global mimeHdr

    MimeInsertInfo $tkw $depth $part
    TextButton $tkw "Pass content to metamail..." \
	[concat [list MimeMetaMail $mimeHdr($depth,$part,hdr,content-type) \
			   $mimeHdr($depth,$part,encoding) \
			   $mimeHdr($depth,$part,file)] \
			   $args]
    return 0
}
proc Mime_ShowDirect { tkw depth part } {
    return [Mime_ShowDefault $tkw $depth $part -e -d]
}
proc MimeInsertInfo { tkw depth part } {
    global mimeHdr
    if [info exists mimeHdr($depth,$part,hdr,content-description)] {
	    $tkw insert insert \
		"Description: $mimeHdr($depth,$part,hdr,content-description) "
    }
    $tkw insert insert "($mimeHdr($depth,$part,type))\n"
    foreach key $mimeHdr($depth,$part,params) {
	$tkw insert insert "\t$key = $mimeHdr($depth,$part,param,$key)\n"
    }
}

proc MimeMetaMail { contentType encoding fileName args } {
    global mimeHdr msgHdr
    if [catch {
	Exmh_Status "metamail $fileName"
	set mcmd [list metamail -b \
		    -c $contentType \
		    -E $encoding \
		    -f [MsgParseFrom $msgHdr(from)] \
		    -m exmh]
	if {[lsearch $args -d] < 0} {
	    set xterm [list xterm -e]
	} else {
	    set xterm {}
	}
	# recall that eval concats its arguments, thus exploding things for us
	eval exec $xterm $mcmd $args $fileName
	Exmh_Status ok
    } err] {
	 Exmh_Status "$err"
    }
}

proc Mime_ShowText { tkw depth part } {
    global mimeHdr
    if [info exists mimeHdr($depth,$part,param,charset)] {
	if ![regexp -nocase us-ascii $mimeHdr($depth,$part,param,charset)] {
	    $tkw insert insert "(Warning: non US-ascii charset <$mimeHdr($depth,$part,param,charset)>)\n\n"
	}	
    }
    set subtype [file tail $mimeHdr($depth,$part,type)]
    set encoding $mimeHdr($depth,$part,encoding)
    set fileName $mimeHdr($depth,$part,file)
    if ![regexp {[78]bit} $encoding] {
	set newFile [Mime_TempFile /tmp/exmh.decode]
	if ![MimeDecode $mimeHdr($depth,$part,file) $newFile $encoding] {
	    $tkw insert insert "(Decode failed - raw text follows)\n\n"
	    set subtype plain
	} else {
	    set fileName $newFile
	}
    }
    Exmh_Status text/$subtype
    if [catch {open $fileName r} fileIO] {
	Exmh_Status "Cannot open body $fileName: $fileIO"
	return 1
    }
    case $subtype {
	default {$tkw insert insert [read $fileIO]}
	plain   {$tkw insert insert [read $fileIO]}
	{enriched richtext} { Rich_Display $tkw $fileIO $subtype}
    }
    close $fileIO
    if [info exists newFile] {
	exec rm $newFile
    }
    return 1
}
proc MimeDecode { fileName name encoding } {
    set ok 1
    if [catch {
	case $encoding {
	    default {
		Exmh_Status "cat > $name"
		exec cat < $fileName > $name
	    }
	    {8bit 7bit} {
		Exmh_Status "cat > $name"
		exec cat < $fileName > $name
	    }
	    base64 {
		Exmh_Status "mimencode -u -b > $name"
		exec mimencode < $fileName -u -b > $name
	    }
	    quoted-printable {
		Exmh_Status "mimencode -u -q > $name"
		exec mimencode < $fileName -u -q > $name
	    }
	}
    } err] {
	Exmh_Status "Decode failed: $err"
	set ok 0
    }
    return $ok
}

proc Mime_ShowApplicationOctet { tkw depth part } {
    global mimeHdr

    $tkw insert insert "You have received an encoded file.\n"
    MimeInsertInfo $tkw $depth $part
    if [info exists mimeHdr($depth,$part,param,name)] {
	set name [string trim $mimeHdr($depth,$part,param,name) \ \" ]
    } else {
	set name {}
    }
    TextButton $tkw "Open file transfer dialog..." \
	[list MimeFileTransfer $mimeHdr($depth,$part,file) $name \
				$mimeHdr($depth,$part,encoding)]
    return 0
}

proc MimeFileTransfer { fileName name encoding } {
    set name [FSBox "Select the name of the downloaded file:" $name ]
    if {$name != {}} {
	if [MimeDecode $fileName $name $encoding] {
	    Exmh_Status "Conversion ok"
	}
    }
}
proc Mime_ShowAudio { tkw depth part } {
    global mimeHdr

    $tkw insert insert "You have received an audio file\n"
    MimeInsertInfo $tkw $depth $part
    TextButton $tkw "Play audio..." \
	[list MimeDoAudio $mimeHdr($depth,$part,file) \
			  $mimeHdr($depth,$part,encoding) $depth $part]
    return 0
}
proc MimeDoAudio { fileName encoding depth part } {
    global mimeHdr
    if [info exists mimeHdr($depth,$part,audio)] {
	Exmh_Status "Playing audio..."
	exec showaudio < $mimeHdr($depth,$part,audio)
	Exmh_Status ""
    } else {
	set name [Mime_TempFile /tmp/exmh.audio]
	if [MimeDecode $fileName $name $encoding] {
	    Exmh_Status "Playing audio..."
	    exec showaudio < $name
	    set mimeHdr($depth,$part,audio) $name
	    Exmh_Status ""
	}
    }
}
proc Mime_ShowMessageExternal { tkw depth part } {
    global mimeHdr

    if [catch {set mimeHdr($depth,$part,param,access-type)} atype] {
	return [Mime_ShowDefault $tkw $depth $part]
    }
    if {$atype != "anon-ftp"} {
	return [Mime_ShowDefault $tkw $depth $part]
    }
    MimeInsertInfo $tkw $depth $part
    TextButton $tkw "Open FTP transfer dialog..." \
	[list MimeFTPTransfer $mimeHdr($depth,$part,param,site) \
			      $mimeHdr($depth,$part,param,directory) \
			      $mimeHdr($depth,$part,param,name) ]
    return 0
}
proc MimeFTPTransfer { site directory theirname } {
    set myname [FSBox "Select the name of the downloaded file:" $theirname ]
    if {$myname != {}} {
	global env
	Exmh_Status "Starting anon-ftp $site"
	set pipe [open "|ftp $site" w]
	puts $pipe "anonymous"
	puts $pipe "$env(USER)@"
	puts $pipe "cd $directory"
	puts $pipe "get $theirname $myname"
	puts $pipe "quit"
	if [catch {close $pipe} err] {
	    if ![regexp -nocase password $err] {
		Exmh_Status $err
	    } else {
		Exmh_Status "FTP transfer complete"
	    }
	} else {
	    Exmh_Status "FTP transfer complete"
	}
    }
}

proc Mime_ShowMultipart { tkw depth part } {
    global mimeHdr

    if ![info exists mimeHdr($depth,$part,param,boundary)] {
	$tkw insert insert "No <boundary> parameter for multipart message\n"
	$tkw insert insert "Raw content follows...\n\n"
	return [Mime_ShowText $tkw $depth $part]
    }
    set fileName $mimeHdr($depth,$part,file)
    set boundary $mimeHdr($depth,$part,param,boundary)
    set type $mimeHdr($depth,$part,type)

    if [catch {open $fileName r} fileIO] {
	$tkw insert insert "Mime_ShowMultipart $fileName: $fileIO\n"
	return 0
    }
    Exmh_Status "Mime_ShowMultipart $depth $part $type"
    set mimeHdr($depth,$part,numParts) \
	[MimeMultiChop $tkw $depth $part $fileIO $boundary]
    close $fileIO

    # Display it

    MimeShowPart $tkw $depth $part

    return 1
}
proc MimeMultiChop {tkw depth part fileIO boundary} {
    global mimeHdr

    set depth $depth-$part
    set part 0
    set state prolog	;# prolog > header > body [ > header > body ] > postlude

    # Read and parse multipart, diverting subparts to files

    while {1} {
	set numBytes [gets $fileIO line]
	if {$numBytes < 0} {
	    return $part
	}
	if {($numBytes == 0) && ($state == "header")} {
	    set state body
	    if [catch {set mimeHdr($depth,$part,hdr,content-type)} contentType] {
		set contentType text/plain
	    }
	    if [catch {set mimeHdr($depth,$part,hdr,content-transfer-encoding)} encoding] {
		set encoding 7bit
	    }
	    set encoding [string trim [string tolower $encoding] \ \" ]
	    set type [MimeHeader $depth $part $contentType $encoding]
	    Exmh_Status "Saving $type ..."
	    if {[file dirname $type] == "multipart"} {
		if ![catch {set mimeHdr($depth,$part,param,boundary)} bndy] {
		    set mimeHdr($depth,$part,numParts) \
			[MimeMultiChop $tkw $depth $part $fileIO $bndy]
		}
	    }
	    continue
	}
	if [regexp ^-- $line] {
	    #Could be a boundary line
	    set trailer [string range $line 2 end]
	    if {[string compare $boundary $trailer] == 0} {
		if {$state != "prolog"} {
		    close $tmpFile
		}
		incr part
		set state header
		set mimeHdr($depth,$part,file) \
		    [Mime_TempFile /tmp/exmh.$depth.$part]
		set tmpFile [open $mimeHdr($depth,$part,file) w]
		Exmh_Status "MimeMultiChop $depth $part $mimeHdr($depth,$part,file)"
		continue
	    } else {
		if {[string compare $boundary-- $trailer] == 0} {
		    close $tmpFile
		    return $part
		}
	    }
	}
	case $state {
	    header {
		set continue [regexp {^[ 	]} $line]
		if {! $continue} {
		    if [regexp -indices {^([^:]+):} $line match hdr] {
			set cur [string tolower \
			    [eval {string range $line} $hdr]]
			set mimeHdr($depth,$part,hdr,$cur) \
			    [string trim \
				[string range $line \
				    [expr [lindex $match 1]+1] end]]
			lappend mimeHdr($depth,$part,hdrs) $cur
		    }
		} else {
		    if [regexp -indices {^[ 	]+} $line match] {
			append mimeHdr($depth,$part,hdr,$cur) \
			    [string range $line \
				    [expr [lindex $match 1]+1] end]
		    }
		}
	    }
	    body {
		# Divert sub-body to a temp file.
		puts $tmpFile $line
	    }
	}
    }
}
proc MimeShowPart { tkw depth part } {
    global mimeHdr
    set type $mimeHdr($depth,$part,type)
    $tkw insert insert --$type\n
    if ![regexp multipart $type] {
	MimeShow $tkw $depth $part
	return
    }
    set numParts $mimeHdr($depth,$part,numParts)
    set depth $depth-$part
    case [file tail $type] {
	default {
	    for {set p 1} {$p <= $numParts} {incr p} {
		MimeShowPart $tkw $depth $p
	    }
	}
	alternative {
	    global mime
	    for {set p 1} {$p <= $numParts} {incr p} {
		set type $mimeHdr($depth,$p,type)
		set hit 0
		foreach t [list $type [file dirname $type]] {
		    if [info exists mime(showproc,$t)] {
			set hit 1
			break
		    }
		}
		if {! $hit} {
		    break
		}
	    }
	    incr p -1
	    MimeShowPart $tkw $depth $p
	}
    }
}
proc Mime_TempFile { basename } {
    global mime
    set uid 0
    while {[file exists $basename.$uid]} {
	incr uid
    }
    lappend mime(junkfiles) $basename.$uid
    return $basename.$uid
}

proc Mime_Debug { args } {
    puts stderr $args
}
